# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import datetime
import json
import shutil
from math import ceil
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from tqdm.auto import tqdm

from pixano.core import Image
from pixano.data.dataset.dataset import Dataset
from pixano.data.dataset.dataset_item import DatasetItem
from pixano.data.exporters.exporter import Exporter
from pixano.data.item.item_object import ItemObject


class COCOExporter(Exporter):
    """Exporter class for COCO instances dataset

    Attributes:
        dataset (Dataset): Dataset to export
        coco_json (dict[str, Any]): Dataset split in COCO format
    """

    dataset: Dataset
    coco_json: dict[str, Any]
    _category_id_count: int  # for dataset without category_id

    def export_dataset(
        self,
        export_dir: Path,
        splits: list[str] = None,
        objects_sources: list[str] = None,
        copy: bool = True,
    ):
        """Export dataset back to original format

        Args:
            export_dir (Path): Export directory
            splits (list[str], optional): Dataset splits to export, all if None. Defaults to None.
            objects_sources (list[str], optional): Objects sources to export, all if None. Defaults to None.
            copy (bool, optional): True to copy files to export directory. Defaults to True.
        """

        # If no splits provided, select all splits
        if splits is None:
            splits = self.dataset.info.splits
            # If no splits, there is nothing to export
            if not splits:
                raise ValueError("Dataset has no splits to export.")

        # If no object sources provided, select all object tables
        if objects_sources is None:
            objects_sources = list(
                table.source for table in self.dataset.info.tables["objects"]
            )
            # If no object tables, there is nothing to export
            if not objects_sources:
                raise ValueError("Dataset has no objects tables to export.")

        # Create export directory
        ann_dir = export_dir / f"annotations [{', '.join(objects_sources)}]"
        ann_dir.mkdir(parents=True, exist_ok=True)

        self._category_id_count = 0  # used if no category_id in dataset (TODO: get from prebuilt coco id/name mapping)

        # Iterate on splits
        with tqdm(desc="Processing dataset", total=self.dataset.num_rows) as progress:
            for split in splits:
                # build categories field from features_values
                if (
                    self.dataset.info.features_values
                    and self.dataset.info.features_values.objects
                    and "category" in self.dataset.info.features_values.objects
                ):
                    if "category_id" in self.dataset.info.features_values.objects:
                        categories = [
                            {"id": id, "name": name, "supercategory": ""}
                            for id, name in zip(
                                self.dataset.info.features_values.objects[
                                    "category_id"
                                ].values,
                                self.dataset.info.features_values.objects[
                                    "category"
                                ].values,
                            )
                        ]
                    else:
                        for i, val in enumerate(
                            self.dataset.info.features_values.objects["category"].values
                        ):
                            categories.append({"id": i, "name": val})

                else:
                    categories = []

                # Create COCO json
                self.coco_json = {
                    "info": {
                        "description": self.dataset.info.name,
                        "url": "N/A",
                        "version": f"v{datetime.datetime.now().strftime('%y%m%d.%H%M%S')}",
                        "year": datetime.date.today().year,
                        "contributor": "Exported from Pixano",
                        "date_created": datetime.date.today().isoformat(),
                    },
                    "licences": [
                        {
                            "url": "N/A",
                            "id": 1,
                            "name": "Unknown",
                        },
                    ],
                    "images": [],
                    "annotations": [],
                    "categories": categories,
                }

                batch_size = 1024

                for batch_index in range(ceil(self.dataset.num_rows / batch_size)):
                    # Load items
                    items = self.dataset.load_items(
                        limit=min(
                            self.dataset.num_rows,
                            (batch_index + 1) * batch_size,
                        ),
                        offset=batch_index * batch_size,
                    )

                    # Iterate on items
                    for item in items:
                        # Filter on split
                        if item.split == split:
                            # Export item
                            self._export_item(item, objects_sources)
                            # Update progress bar
                            progress.update(1)

                # Save COCO format .json file
                with open(
                    ann_dir / f"instances_{split}.json", "w", encoding="utf-8"
                ) as f:
                    json.dump(self.coco_json, f)

        # Copy media directory
        if copy:
            if (
                self.dataset.media_dir.exists()
                and self.dataset.media_dir != export_dir / "media"
            ):
                shutil.copytree(
                    self.dataset.media_dir, export_dir / "media", dirs_exist_ok=True
                )

    def _export_item(
        self, item: DatasetItem, objects_sources: list[str]
    ) -> dict[str, Image]:
        """Export item to COCO format

        Args:
            item (DatasetItem): Item to export
            objects_sources (list[str]): Object sources to export

        Returns:
            dict[str, Image]: Loaded images
        """

        # Export images
        images = self._export_images(item)
        # Export objects
        item_with_objects = self.dataset.load_item(item.id, load_objects=True)
        for obj in item_with_objects.objects.values():
            # Filter by views and object sources
            if obj.view_id in images and obj.source_id in objects_sources:
                # Export object
                self._export_object(obj, item, images)

    def _export_images(self, item: DatasetItem) -> dict[str, Image]:
        """Export item images to COCO format

        Args:
            item (DatasetItem): Item

        Returns:
            dict[str, Image]: Loaded images
        """

        images: dict[str, Image] = {}
        for view in item.views.values():
            if view.type == "image":
                # Reformat URI for export
                export_uri = (
                    view.uri.replace(f"data/{self.dataset.path.name}/media/", "")
                    if urlparse(view.uri).scheme == ""
                    else view.uri
                )
                # Create image from URI
                images[view.id] = Image(
                    uri=export_uri,
                    uri_prefix=self.dataset.media_dir.absolute().as_uri(),
                )
                # Append image info
                self.coco_json["images"].append(
                    {
                        "license": 1,
                        "coco_url": images[view.id].complete_uri,
                        "file_name": images[view.id].file_name,
                        "height": images[view.id].height,
                        "width": images[view.id].width,
                        "id": item.original_id if item.original_id else item.id,
                    }
                )

        return images

    def _export_object(
        self, obj: ItemObject, item: DatasetItem, images: dict[str, Image]
    ) -> dict[str, Any]:
        """Export item object to COCO format

        Args:
            obj (ItemObject): Object to export
            item (DatasetItem): Item
            images (dict[str, Image]): Item images

        Returns:
            dict[str, Any]: Object in COCO format
        """

        # Bounding box
        bbox = (
            obj.bbox.to_pyarrow()
            .denormalize(
                height=images[obj.view_id].height,
                width=images[obj.view_id].width,
            )
            .xywh_coords
            if obj.bbox
            else None
        )

        # Mask
        mask = obj.mask.to_pyarrow().to_urle() if obj.mask else None

        # Add to categories
        category = {}
        if "category" in obj.features and obj.features["category"].value not in [
            cat["name"] for cat in self.coco_json["categories"]
        ]:
            if "category_id" in obj.features:
                category["id"] = obj.features["category_id"].value
            else:
                category["id"] = self._category_id_count
                self._category_id_count += 1
            category["name"] = obj.features["category"].value
            category["supercategory"] = (
                obj.features["supercategory"].value
                if "supercategory" in obj.features
                else ""
            )
            self.coco_json["categories"].append(category)
        else:
            category = {
                "id": [
                    cat["id"]
                    for cat in self.coco_json["categories"]
                    if cat["name"] == obj.features["category"].value
                ][0],
                "name": obj.features["category"].value,
            }
            if "supercategory" in obj.features:
                category["supercategory"] = obj.features["supercategory"].value
                for cat in self.coco_json["categories"]:
                    if cat["name"] == category["name"]:
                        cat["supercategory"] = category["supercategory"]
            else:
                for cat in self.coco_json["categories"]:
                    if cat["name"] == category["name"]:
                        cat["supercategory"] = ""

        # Add object
        ann_object = {
            "id": obj.original_id if obj.original_id else obj.id,
            "image_id": item.original_id if item.original_id else item.id,
            "segmentation": mask,
            "bbox": bbox,
            "area": 0,
            "iscrowd": 0,
            "category_id": category["id"],
            "category_name": category["name"],
        }
        if category["supercategory"] != "":
            ann_object["supercategory"] = category["supercategory"]
        self.coco_json["annotations"].append(ann_object)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
from pathlib import Path
from typing import Optional

from pydantic import BaseModel
from s3path import S3Path

from pixano.core import Image
from pixano.data.dataset.dataset_stat import DatasetStat
from pixano.data.dataset.dataset_table import DatasetTable
from pixano.data.item.item_feature import FeaturesValues


class DatasetInfo(BaseModel):
    """DatasetInfo

    Attributes:
        id (str): Dataset ID
        name (str): Dataset name
        description (str): Dataset description
        estimated_size (str): Dataset estimated size
        num_elements (int): Number of elements in dataset
        splits (list[str]): Dataset splits
        tables (dict[str, list[DatasetTable]]): Dataset tables
        features_values: (FeaturesValues, optional): existing values for each custom feature
        preview (str, optional): Dataset preview
        stats (list[DatasetStat], optional): Dataset stats
    """

    id: str
    name: str
    description: str
    estimated_size: str
    num_elements: int
    splits: list[str]
    tables: dict[str, list[DatasetTable]]
    features_values: Optional[FeaturesValues] = None
    preview: Optional[str] = None
    stats: Optional[list[DatasetStat]] = None

    def save(self, save_dir: Path | S3Path):
        """Save DatasetInfo to json file

        Args:
            save_dir (Path | S3Path): Save directory
        """
        if isinstance(save_dir, S3Path):
            with (save_dir / "db.json").open(encoding="utf-8") as f:
                json.dump(self.model_dump(), f)
        else:
            with open(save_dir / "db.json", "w", encoding="utf-8") as f:
                json.dump(self.model_dump(), f)

    @staticmethod
    def from_json(
        json_fp: Path | S3Path,
        load_stats: bool = False,
        load_thumbnail: bool = False,
    ) -> "DatasetInfo":
        """Read DatasetInfo from JSON file

        Args:
            json_fp (Path | S3Path): JSON file path
            load_stats (bool, optional): Load dataset stats. Defaults to False.
            load_thumbnail (bool, optional): Load dataset thumbnail. Defaults to False.

        Returns:
            DatasetInfo: DatasetInfo
        """

        if isinstance(json_fp, S3Path):
            with json_fp.open(encoding="utf-8") as json_file:
                info_json = json.load(json_file)
        else:
            with open(json_fp, encoding="utf-8") as json_file:
                info_json = json.load(json_file)

        info = DatasetInfo.model_validate(info_json)

        # Load dataset stats file
        if load_stats:
            stats_fp = json_fp.parent / "stats.json"
            if stats_fp.is_file():
                info.stats = DatasetStat.from_json(stats_fp)

        # Load thumbnail
        if load_thumbnail:
            thumb_fp = json_fp.parent / "preview.png"
            if thumb_fp.is_file():
                if isinstance(json_fp, S3Path):
                    info.preview = thumb_fp.get_presigned_url()
                else:
                    im = Image(uri=thumb_fp.absolute().as_uri())
                    info.preview = im.url

        return info

    @staticmethod
    def load_directory(
        directory: Path | S3Path,
        load_thumbnail: bool = False,
        load_stats: bool = False,
    ) -> list["DatasetInfo"]:
        """Load list of DatasetInfo from directory

        Args:
            directory (Path | S3Path): Directory to load
            load_thumbnail (bool, optional): Load dataset thumbnail. Defaults to False.
            load_stats (bool, optional): Load dataset stats. Defaults to False.

        Returns:
            list[DatasetInfo]: List of DatasetInfo
        """

        infos = []

        # Browse directory
        for json_fp in sorted(directory.glob("*/db.json")):
            # Add dataset info to list
            infos.append(
                DatasetInfo.from_json(
                    json_fp,
                    load_thumbnail=load_thumbnail,
                    load_stats=load_stats,
                )
            )

        return infos



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purcompressedRLE is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import numpy as np
import pyarrow as pa
from PIL import Image
from pydantic import BaseModel, PrivateAttr

from pixano.core.pixano_type import PixanoType, create_pyarrow_type
from pixano.utils import (
    encode_rle,
    mask_to_rle,
    polygons_to_rle,
    rle_to_mask,
    rle_to_polygons,
    rle_to_urle,
    urle_to_rle,
)


class CompressedRLE(PixanoType, BaseModel):
    """Compressed RLE mask type

    Attributes:
        _size (list[int]): Mask size
        _counts (bytes): Mask RLE encoding
    """

    _size: list[int] = PrivateAttr()
    _counts: bytes = PrivateAttr()

    def __init__(self, size: list[int], counts: bytes):
        """Initalize compressed RLE mask

        Args:
            size (list[int]): Mask size
            counts (bytes): Mask RLE encoding
        """

        # Define public attributes through Pydantic BaseModel
        super().__init__()

        # Define private attributes manually
        self._size = size
        self._counts = counts

    @property
    def size(self) -> list[int]:
        """Return mask size

        Returns:
            list[int]: Mask size
        """

        return self._size

    @property
    def counts(self) -> bytes:
        """Return mask RLE encoding

        Returns:
            bytes: Mask RLE encoding
        """

        return self._counts if not None else b""

    def to_mask(self) -> np.ndarray:
        """Convert compressed RLE mask to NumPy array

        Returns:
            np.ndarray: Mask as NumPy array
        """

        return rle_to_mask(self.to_dict())

    def to_urle(self) -> dict[str, list[int]]:
        """Convert compressed RLE mask to uncompressed RLE

        Returns:
            dict[str, list[int]]: Mask as uncompressed RLE
        """

        return rle_to_urle(self.to_dict())

    def to_polygons(self) -> list[list]:
        """Convert compressed RLE mask to poylgons

        Returns:
            list[list]: Mask as polygons
        """

        return rle_to_polygons(self.to_dict())

    @staticmethod
    def from_mask(mask: Image.Image | np.ndarray) -> "CompressedRLE":
        """Create compressed RLE mask from NumPy array

        Args:
            mask (Image.Image | np.ndarray): Mask as NumPy array

        Returns:
            CompressedRLE: Compressed RLE mask
        """

        return CompressedRLE.from_dict(mask_to_rle(mask))

    @staticmethod
    def from_urle(urle: dict[str, list[int]]) -> "CompressedRLE":
        """Create compressed RLE mask from uncompressed RLE

        Args:
            urle (dict[str, list[int]]): Mask as uncompressed RLE

        Returns:
            CompressedRLE: Compressed RLE mask
        """

        return CompressedRLE.from_dict(urle_to_rle(urle))

    @staticmethod
    def from_polygons(
        polygons: list[list],
        height: int,
        width: int,
    ) -> "CompressedRLE":
        """Create compressed RLE mask from polygons

        Args:
            polygons (list[list]): Mask as polygons
            height (int): Image height
            width (int): Image width

        Returns:
            CompressedRLE: Compressed RLE mask
        """

        return CompressedRLE.from_dict(polygons_to_rle(polygons, height, width))

    @staticmethod
    def encode(
        mask: list[list] | dict[str, list[int]], height: int, width: int
    ) -> "CompressedRLE":
        """Create compressed RLE mask from polygons / uncompressed RLE / compressed RLE

        Args:
            mask (list[list] | dict[str, list[int]]): Mask as polygons / uncompressed RLE / compressed RLE
            height (int): Image height
            width (int): Image width

        Returns:
            CompressedRLE: Compressed RLE mask
        """

        return CompressedRLE.from_dict(encode_rle(mask, height, width))

    @staticmethod
    def to_struct() -> pa.StructType:
        """Return CompressedRLE type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """

        return pa.struct(
            [
                pa.field("size", pa.list_(pa.int32(), list_size=2)),
                pa.field("counts", pa.binary()),
            ]
        )


CompressedRLEType = create_pyarrow_type(
    CompressedRLE.to_struct(), "CompressedRLE", CompressedRLE
)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pathlib import Path
from typing import Any, Optional

import lancedb
import pyarrow as pa
from pydantic import BaseModel

from pixano.data.dataset.dataset_info import DatasetInfo
from pixano.data.fields import Fields, field_to_python
from pixano.data.item import ItemEmbedding, ItemFeature, ItemObject, ItemView


class DatasetItem(BaseModel):
    """DatasetItem

    Attributes:
        id (str): Item ID
        original_id (str, optional): Item original ID
        split (str): Item split
        features (dict[str, ItemFeature], optional): Item features
        views (dict[str, ItemView], optional): Item views
        objects (dict[str, ItemObject], optional): Item objects
        embeddings (dict[str, ItemEmbedding], optional): Item embeddings
    """

    id: str
    original_id: Optional[str] = None
    split: str
    features: Optional[dict[str, ItemFeature]] = None
    views: Optional[dict[str, ItemView]] = None
    objects: Optional[dict[str, ItemObject]] = None
    embeddings: Optional[dict[str, ItemEmbedding]] = None

    @staticmethod
    def from_pyarrow(
        pyarrow_item: dict[str, dict[str, pa.Table]],
        info: DatasetInfo,
        media_dir: Path,
        media_features: bool = False,
        model_id: str = None,
    ) -> "DatasetItem":
        """Format PyArrow item

        Args:
            pyarrow_item (dict[str, dict[str, pa.Table]]): PyArrow item
            info (DatasetInfo): Dataset info
            media_dir (Path): Dataset media directory
            media_features (bool, optional): Load media features like image width and height (slow for large item batches)
            model_id (str, optional): Model ID (ONNX file path) of embeddings to load. Defaults to None.

        Returns:
            DatasetItem: Formatted item
        """

        item_info = pyarrow_item["main"]["db"].to_pylist()[0]

        # Create item
        item = DatasetItem(
            id=item_info["id"],
            split=item_info["split"],
        )

        for group_name, table_group in info.tables.items():
            # Main table
            if group_name == "main":
                # Item features
                item.features = ItemFeature.from_pyarrow(
                    pyarrow_item["main"]["db"],
                    Fields(table_group[0].fields).to_schema(),
                )

            # Media tables
            if group_name == "media" and "media" in pyarrow_item:
                item.views = {}
                for table in table_group:
                    item.views = item.views | ItemView.from_pyarrow(
                        pyarrow_item["media"][table.name],
                        Fields(table.fields).to_schema(),
                        media_dir,
                        media_features,
                    )

            # Objects
            if group_name == "objects" and "objects" in pyarrow_item:
                item.objects = {}
                for table in table_group:
                    item.objects = item.objects | ItemObject.from_pyarrow(
                        pyarrow_item["objects"][table.name],
                        Fields(table.fields).to_schema(),
                        table.source,
                    )

            # Active Learning
            if group_name == "active_learning" and "active_learning" in pyarrow_item:
                for table in table_group:
                    al_features = ItemFeature.from_pyarrow(
                        pyarrow_item["active_learning"][table.name],
                        Fields(table.fields).to_schema(),
                    )
                    item.features = item.features | al_features

            # Segmentation embeddings
            if group_name == "embeddings" and "embeddings" in pyarrow_item:
                item.embeddings = {}
                for table in table_group:
                    if table.source.lower() in model_id.lower():
                        item.embeddings = item.embeddings | ItemEmbedding.from_pyarrow(
                            pyarrow_item["embeddings"][table.name],
                            Fields(table.fields).to_schema(),
                        )

        return item

    def to_pyarrow(self) -> dict[str, Any]:
        """Return DatasetItem in PyArrow format

        Returns:
            dict[str, Any]: Item in PyArrow format
        """

        pyarrow_item = {}

        # ID
        pyarrow_item["id"] = self.id
        pyarrow_item["split"] = self.split

        # Features
        if self.features is not None:
            # Add features
            for feat in self.features.values():
                pyarrow_item[feat.name] = (
                    field_to_python(feat.dtype)(feat.value)
                    if feat.value is not None
                    else None
                )

            # Check feature types
            for feat in self.features.values():
                if pyarrow_item[feat.name] is not None and not isinstance(
                    pyarrow_item[feat.name], field_to_python(feat.dtype)
                ):
                    raise ValueError(
                        f"Feature {feat.name} of object {self.id} is of type {type(self.features[feat.name].value)} instead of type {field_to_python(feat.dtype)}"
                    )

        return pyarrow_item

    def update(
        self,
        ds_table: lancedb.db.LanceTable,
    ):
        """Update dataset item

        Args:
            ds_table (lancedb.db.LanceTable): Item table
        """

        # Convert item to PyArrow
        pyarrow_item = self.to_pyarrow()
        table_item = pa.Table.from_pylist(
            [pyarrow_item],
            schema=ds_table.schema,
        )

        # Update item
        ds_table.delete(f"id in ('{self.id}')")
        ds_table.add(table_item, mode="append")

        # Clear change history to prevent dataset from becoming too large
        ds_table.to_lance().cleanup_old_versions()

    def delete_objects(
        self,
        ds_tables: dict[str, dict[str, lancedb.db.LanceTable]],
    ):
        """Delete remove objects from dataset item

        Args:
            ds_tables (dict[str, dict[str, lancedb.db.LanceTable]]): Dataset tables
        """

        # Get current item objects
        current_obj_tables = {}
        for table_name, table in ds_tables["objects"].items():
            media_scanner = table.to_lance().scanner(filter=f"item_id in ('{self.id}')")
            current_obj_tables[table_name] = media_scanner.to_table().to_pylist()

        # Check if objects have been deleted
        for table_name, current_obj_table in current_obj_tables.items():
            for current_obj in current_obj_table:
                # If object has been deleted
                if not any(
                    obj_id == current_obj["id"] for obj_id in self.objects.keys()
                ):
                    # Remove object from table
                    ds_tables["objects"][table_name].delete(
                        f"id in ('{current_obj['id']}')"
                    )

                    # Clear change history to prevent dataset from becoming too large
                    ds_tables["objects"][table_name].to_lance().cleanup_old_versions()



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pixano.core.bbox import BBox, BBoxType
from pixano.core.camera import Camera, CameraType
from pixano.core.compressed_rle import CompressedRLE, CompressedRLEType
from pixano.core.depth_image import DepthImage, DepthImageType
from pixano.core.gt_info import GtInfo, GtInfoType
from pixano.core.image import Image, ImageType
from pixano.core.pixano_type import PixanoType, convert_field, create_pyarrow_type
from pixano.core.pose import Pose, PoseType
from pixano.core.utils import (
    is_binary,
    is_boolean,
    is_float,
    is_image_type,
    is_integer,
    is_string,
)

__all__ = [
    "BBox",
    "BBoxType",
    "Camera",
    "CameraType",
    "CompressedRLE",
    "CompressedRLEType",
    "DepthImage",
    "DepthImageType",
    "GtInfo",
    "GtInfoType",
    "Image",
    "ImageType",
    "PixanoType",
    "convert_field",
    "create_pyarrow_type",
    "Pose",
    "PoseType",
    "is_binary",
    "is_boolean",
    "is_float",
    "is_image_type",
    "is_integer",
    "is_string",
]



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info


import json
from pathlib import Path
from typing import Optional

from pydantic import BaseModel
from s3path import S3Path


class DatasetStat(BaseModel):
    """DatasetStat

    Attributes:
        name (str): Stats name
        type (str): Stats type ('numerical' or 'categorical')
        histogram (str): Stats histogram data
    """

    name: str
    type: str
    histogram: list[dict[str, float | int | str]]
    range: Optional[list[int | float]] = None

    @staticmethod
    def from_json(json_fp: Path | S3Path) -> list["DatasetStat"]:
        """Read list of DatasetStats from JSON file

        Args:
            json_fp (Path | S3Path): JSON file path

        Returns:
            list[DatasetStats]: List of DatasetStat
        """

        if isinstance(json_fp, S3Path):
            with json_fp.open(encoding="utf-8") as json_file:
                stats_json = json.load(json_file)
        else:
            with open(json_fp, encoding="utf-8") as json_file:
                stats_json = json.load(json_file)

        return [DatasetStat.model_validate(stat) for stat in stats_json]

    def save(self, save_dir: Path | S3Path):
        """Save DatasetInfo to json file
           replace existing histogram with same name in json_fp

        Args:
            save_dir (Path | S3Path): Save directory
        """

        try:
            if isinstance(save_dir, S3Path):
                with (save_dir / "stats.json").open(encoding="utf-8") as json_file:
                    json_stats = json.load(json_file)
            else:
                with open(save_dir / "stats.json", "r", encoding="utf-8") as json_file:
                    json_stats = json.load(json_file)
        except FileNotFoundError:
            json_stats = []
        # keep all stats except the one with same name, we replace it if exist
        json_stats = [stat for stat in json_stats if stat["name"] != self.name]
        json_stats.append(
            {"name": self.name, "type": self.type, "histogram": self.histogram}
        )

        if isinstance(save_dir, S3Path):
            with (save_dir / "stats.json").open("w", encoding="utf-8") as f:
                json.dump(json_stats, f, indent="\t")
        else:
            with open(save_dir / "stats.json", "w", encoding="utf-8") as f:
                json.dump(json_stats, f, indent="\t")



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException

from pixano.data import Dataset, DatasetInfo, Settings, get_settings

router = APIRouter(tags=["datasets"])


@router.get("/datasets", response_model=list[DatasetInfo])
async def get_datasets(
    settings: Annotated[Settings, Depends(get_settings)]
) -> list[DatasetInfo]:
    """Load dataset list

    Args:
        settings (Settings): App settings

    Returns:
        list[DatasetInfo]: List of dataset infos
    """

    # Load datasets
    infos = DatasetInfo.load_directory(directory=settings.data_dir, load_thumbnail=True)

    # Return datasets
    if infos:
        return infos
    raise HTTPException(
        status_code=404,
        detail=f"No datasets found in {settings.data_dir.absolute()}",
    )


@router.get("/datasets/{ds_id}", response_model=DatasetInfo)
async def get_dataset(
    ds_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> DatasetInfo:
    """Load dataset

    Args:
        ds_id (str): Dataset ID
        settings (Settings): App settings

    Returns:
        DatasetInfo: Dataset info
    """

    # Load dataset
    dataset = Dataset.find(ds_id, settings.data_dir)

    # Return dataset info
    if dataset:
        return dataset.load_info(load_stats=True, load_features_values=True)
    raise HTTPException(
        status_code=404,
        detail=f"Dataset {ds_id} not found in {settings.data_dir.absolute()}",
    )



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pixano.data.dataset import (
    Dataset,
    DatasetInfo,
    DatasetItem,
    DatasetStat,
    DatasetTable,
)
from pixano.data.exporters import COCOExporter, Exporter
from pixano.data.fields import Fields, field_to_pyarrow, field_to_python
from pixano.data.importers import COCOImporter, DOTAImporter, ImageImporter, Importer
from pixano.data.item import ItemEmbedding, ItemFeature, ItemObject, ItemView
from pixano.data.settings import Settings, get_settings

__all__ = [
    "Dataset",
    "DatasetInfo",
    "DatasetItem",
    "DatasetStat",
    "DatasetTable",
    "ItemEmbedding",
    "ItemFeature",
    "ItemObject",
    "ItemView",
    "Fields",
    "field_to_pyarrow",
    "field_to_python",
    "Settings",
    "get_settings",
    "Exporter",
    "COCOExporter",
    "Importer",
    "ImageImporter",
    "DOTAImporter",
    "COCOImporter",
]



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import glob
from collections.abc import Iterator
from pathlib import Path

import shortuuid

from pixano.core import Image
from pixano.data.importers.importer import Importer
from pixano.utils import image_to_thumbnail, natural_key


class ImageImporter(Importer):
    """Importer class for image datasets

    Attributes:
        info (DatasetInfo): Dataset information
        input_dirs (dict[str, Path]): Dataset input directories
    """

    def __init__(
        self,
        name: str,
        description: str,
        input_dirs: dict[str, Path],
        splits: list[str] = None,
        media_fields: dict[str, str] = None,
    ):
        """Initialize Image Importer

        Args:
            name (str): Dataset name
            description (str): Dataset description
            input_dirs (dict[str, Path]): Dataset input directories
            splits (list[str], optional): Dataset splits. Defaults to None for datasets with no subfolders for splits.
            media_fields (dict[str, str]): Dataset media fields, with field names as keys and field types as values. Default to None.
        """

        # Create dataset tables
        tables = super().create_tables(media_fields)

        # Create splits
        if splits is None:
            splits = ["dataset"]

        # Initialize Importer
        self.input_dirs = input_dirs
        super().__init__(name, description, tables, splits)

    def import_rows(self) -> Iterator:
        """Process dataset rows for import

        Yields:
            Iterator: Processed rows
        """

        for split in self.info.splits:
            # Get images paths
            image_paths = []
            for ftype in ["*.png", "*.jpg", "*.jpeg"]:
                if split == "dataset":
                    image_paths.extend(glob.glob(str(self.input_dirs["image"] / ftype)))
                else:
                    image_paths.extend(
                        glob.glob(str(self.input_dirs["image"] / split / ftype))
                    )
            image_paths = [Path(p) for p in sorted(image_paths, key=natural_key)]

            # Process rows
            for im_path in image_paths:
                # Create image thumbnail
                im_thumb = image_to_thumbnail(im_path.read_bytes())

                # Set image URI
                im_uri = (
                    f"image/{im_path.name}"
                    if split == "dataset"
                    else f"image/{split}/{im_path.name}"
                )

                # Set unique item id
                item_id = shortuuid.uuid()

                # Return rows
                rows = {
                    "main": {
                        "db": [
                            {
                                "id": item_id,
                                "original_id": im_path.name,
                                "views": ["image"],
                                "split": split,
                                "label": "",
                            }
                        ]
                    },
                    "media": {
                        "image": [
                            {
                                "id": item_id,
                                "image": Image(im_uri, None, im_thumb).to_dict(),
                            }
                        ]
                    },
                }
                yield rows
